@@ -1,6 +1,5 @@
 
 from IPython.display import Markdown, display, Image, SVG, HTML
-import matplotlib.pyplot as plt
 import numpy as np
 from collections import defaultdict
 from contextlib import redirect_stdout, redirect_stderr
@@ -15,6 +14,12 @@ import pandas as pd
 from math import log10
 import shutil
 
+import matplotlib
+from matplotlib.patches import Rectangle, Polygon
+import matplotlib.pyplot as plt
+from matplotlib_inline.backend_inline import set_matplotlib_formats
+set_matplotlib_formats('retina', 'png')
+
 from goatools.base import download_go_basic_obo
 #from goatools.base import download_ncbi_associations
 from goatools.obo_parser import GODag, OBOReader
@@ -33,11 +38,59 @@ from Bio import Entrez
 
 from .intervals import *
 
+# sys.path.append('.')
+
 CACHE = dict()
 
+class NotFound(Exception):
+    """query returned no result"""
+    pass
+
+
+def ensembl_id(name, species='homo_sapiens'):
+    server = "https://rest.ensembl.org"
+    ext = f"/xrefs/symbol/{species}/{name}?"
+    r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
+    if not r.ok:
+      r.raise_for_status()
+    decoded = r.json()
+    ensembl_ids = [x['id'] for x in decoded if x['type'] == 'gene']
+    if not len(ensembl_ids) == 1:
+        raise NotFound
+    return ensembl_ids[0]
+
+def ensembl2symbol(ensembl_id):
+    server = "https://rest.ensembl.org"
+    ext = f"/xrefs/id/{ensembl_id}?"
+    r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
+    if not r.ok:
+        r.raise_for_status()
+    decoded = r.json()
+    symbols = [x['display_id'] for x in decoded if x['dbname'] == 'HGNC']
+    if not len(symbols) == 1:
+        raise NotFound
+    return symbols[0]
+
+def hgcn_symbol(name):
+    if type(name) is list or type(name) is set:
+        return [ensembl2symbol(ensembl_id(n)) for n in name]
+    else:
+        return ensembl2symbol(ensembl_id(name))
+
+def ensembl2ncbi(ensembl_id):
+    server = "https://rest.ensembl.org"
+    ext = f"/xrefs/id/{ensembl_id}?"
+    r = requests.get(server+ext, headers={ "Content-Type" : "application/json"})
+    if not r.ok:
+      r.raise_for_status()
+    decoded = r.json()
+    ids = [x['primary_id'] for x in decoded if x['dbname'] == 'EntrezGene']
+    if not len(ids) == 1:
+        raise NotFound
+    return int(ids[0])
 
 def mygene_get_gene_info(query, species='human', scopes='hgnc', fields='symbol,alias,name,type_of_gene,summary,genomic_pos,genomic_pos_hg19'):
-    api_url = f"https://mygene.info/v3/query?content-type=appliation/x-www-form-urlencoded;q={query};scopes={scopes};species={species};fields={fields}"
+    api_url = f"https://mygene.info/v3/query?q={query}&scopes={scopes}&species={species}&fields={fields}"    
     response = requests.get(api_url)
     if not response.ok:
         response.raise_for_status()
@@ -56,8 +109,15 @@ def gene_info(query, species='human', scopes='hgnc'):
         
     for gene in query:
 
-        top_hit = mygene_get_gene_info(gene, species=species, scopes=scopes,
-                           fields='symbol,alias,name,type_of_gene,summary,genomic_pos,genomic_pos_hg19')
+        for i in range(3):
+            try:
+                top_hit = mygene_get_gene_info(gene, species=species, scopes=scopes,
+                                fields='symbol,alias,name,type_of_gene,summary,genomic_pos,genomic_pos_hg19')
+            except KeyError:
+                continue
+            else:
+                break
+
 
         tmpl = "**Symbol:** **_{symbol}_** "
 
@@ -96,7 +156,7 @@ def gene_info(query, species='human', scopes='hgnc'):
         display(Markdown(tmpl.format(**top_hit)))
 
 
-def _ensembl_get_features_region(chrom, window_start, window_end, features=['gene', 'exon'], assembly=None, species='human'):
+def _ensembl_get_features_region(chrom, window_start, window_end, features=['gene', 'exon'], assembly=None, species='homo_sapiens'):
     if chrom.startswith('chr'):
         chrom = chrom[3:]
     window_start, window_end = int(window_start), int(window_end)
@@ -134,11 +194,10 @@ def ensembl_get_gene_info_by_symbol(symbols, assembly=None, species='homo_sapien
     r = requests.post(server+ext, headers=headers, data=f'{{ "symbols": {json.dumps(symbols)} }}')
     if not r.ok:
         r.raise_for_status()
-        sys.exit()     
     return r.json()
 
 
-def ensembl_get_genes_region(chrom, window_start, window_end, assembly=None, species='human'):
+def ensembl_get_genes_region(chrom, window_start, window_end, assembly=None, species='homo_sapiens'):
     
     gene_info = _ensembl_get_features_region(chrom, window_start, window_end, features=['gene'], assembly=assembly, species=species)
     exon_info = _ensembl_get_features_region(chrom, window_start, window_end, features=['exon'], assembly=assembly, species=species)
@@ -157,55 +216,49 @@ def ensembl_get_genes_region(chrom, window_start, window_end, assembly=None, spe
     return gene_info
 
 
-def get_genes_region(chrom, window_start, window_end, only_protein_coding=True, hg19=False, species='human'):
+def get_genes_region(chrom, window_start, window_end, assembly='GRCh38', db='ncbiRefSeq'):
 
-    if hg19:
-        assembly='GRCh37'
-    else:
-        assembly=None
+    
+    api_url = f'https://api.genome.ucsc.edu/getData/track'
+    params = {'track': db,
+              'genome': assembly,
+              'chrom': chrom,
+              'start': window_start,
+              'end': window_end
+              }
+    response = requests.get(api_url, data=params)
+    if not response.ok:
+        response.raise_for_status()
 
     genes = []
-    gene_info = ensembl_get_genes_region(chrom, window_start, window_end, assembly=assembly, species=species)
-    for gene in gene_info.values():
+    for gene in response.json()[db]:
+        exon_starts = [int(x) for x in gene['exonStarts'].split(',') if x]
+        exon_ends = [int(x) for x in gene['exonEnds'].split(',') if x]
+        exons = list(zip(exon_starts, exon_ends))
+        genes.append((gene['name2'], gene['txStart'], gene['txEnd'], gene['strand'], exons))
 
-
-        if gene['biotype'] != 'protein_coding' and only_protein_coding:
-            continue
-
-        if 'external_name' in gene:
-            name = gene['external_name']
-        else:
-            name = gene['id']
-        genes.append((name, gene['start'], gene['end'], gene['strand'], gene['exons'], gene['biotype']))
     return genes
 
 
-def gene_info_region(chrom, window_start, window_end, only_protein_coding=True, hg19=False, species='human'):
-    # mg.query(f'q={chrom}:{start}-{end}', species='human', fetch_all=True):
-    
-    for gene in get_genes_region(chrom, window_start, window_end, 
-                                    only_protein_coding=only_protein_coding, hg19=hg19, species=species):
+def gene_info_region(chrom, window_start, window_end, assembly='GRCh38', db='ncbiRefSeq'):
+
+    for gene in get_genes_region(chrom, window_start, window_end, assembly, db):    
         gene_info(gene[0])
 
 
-def get_genes_region_dataframe(chrom, start, end, hg19=False):
+def get_genes_region_dataframe(chrom, start, end, assembly='GRCh38', db='ncbiRefSeq'):
     try:
         import pandas as pd
     except ImportError:
         print("pandas must be installed to return data frame")
         return
-    genes = get_genes_region(chrom, start, end, hg19=hg19)
+    genes = get_genes_region(chrom, start, end, assembly, db)
     return pd.DataFrame().from_records([x[:4] for x in genes], columns=['name', 'start', 'end', 'strand'])
 
 
-def _plot_gene(name, txstart, txend, strand, exons, gene_type, offset, line_width, min_visible_width, font_size, ax, highlight=False, clip_on=True):
+def _plot_gene(name, txstart, txend, strand, exons, offset, line_width, min_visible_width, font_size, ax, highlight=False, clip_on=True):
 
-    if gene_type == 'protein_coding':
-        color='black'
-    elif gene_type == 'ncrna':
-        color='blue'
-    else:
-        color='green'
+    color='black'
 
     line = ax.plot([txstart, txend], [offset, offset], color=color, linewidth=line_width/5, alpha=0.5)
     line[0].set_solid_capstyle('butt')
@@ -228,22 +281,31 @@ def _plot_gene(name, txstart, txend, strand, exons, gene_type, offset, line_widt
             fontsize=font_size, color=color, clip_on=clip_on)
 
 
-def gene_plot(chrom, start, end, highlight=[], hg19=False, only_protein_coding=False, hard_limits=False, exact_exons=False, figsize=None, clip_on=True):
+def gene_plot(chrom, start, end, assembly, highlight=[], db='ncbiRefSeq', 
+                collapse_splice_var=True, hard_limits=False, exact_exons=False, 
+                figsize=None, aspect=1, despine=False, clip_on=True):
     
     global CACHE
 
-    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize, sharex='col', sharey='row')
+    fig, (ax1, ax2) = plt.subplots(2, 1, figsize=figsize, sharex='col', 
+                                    sharey='row', gridspec_kw={'height_ratios': [1, aspect]})
     plt.subplots_adjust(wspace=0, hspace=0.05)
 
-    if (chrom, start, end, hg19) in CACHE:
-        genes = CACHE[(chrom, start, end, only_protein_coding, hg19)]
+    if (chrom, start, end, assembly) in CACHE:
+        genes = CACHE[(chrom, start, end, assembly)]
     else:
-        genes = list(get_genes_region(chrom, start, end, only_protein_coding=only_protein_coding, hg19=hg19))
-        CACHE[(chrom, start, end, only_protein_coding, hg19)] = genes
+        genes = list(get_genes_region(chrom, start, end, assembly, db))
+        CACHE[(chrom, start, end, assembly)] = genes
+
+    if collapse_splice_var:
+        d = {}
+        for name, txstart, txend, strand, exons in genes:
+            if name not in d:
+                d[name] = [name, txstart, txend, strand, set(exons)]
+            else:
+                d[name][-1].update(exons)
+        genes = d.values()
 
-    if len(genes) == 200:
-        print("Limit reached, make window smaller")
-        return
 
     line_width = max(6, int(50 / log10(end - start)))-2
     font_size = max(6, int(50 / log10(end - start)))
@@ -254,10 +316,10 @@ def gene_plot(chrom, start, end, highlight=[], hg19=False, only_protein_coding=F
         min_visible_exon_width = (end - start) / 1000
         
     plotted_intervals = defaultdict(list)
-    for name, txstart, txend, strand, exons, gene_type in genes:
+    for name, txstart, txend, strand, exons in genes:
 
         gene_interval = [txstart-label_width, txend]
-        max_gene_rows = 400
+        max_gene_rows = 1000
         for offset in range(1, max_gene_rows, 1):
             if not intersect([gene_interval], plotted_intervals[offset]) and \
                 not intersect([gene_interval], plotted_intervals[offset-1]) and \
@@ -279,7 +341,7 @@ def gene_plot(chrom, start, end, highlight=[], hg19=False, only_protein_coding=F
         else:
             hl = None
 
-        _plot_gene(name, txstart, txend, strand, exons, gene_type, 
+        _plot_gene(name, txstart, txend, strand, exons, 
                   offset, line_width, min_visible_exon_width, font_size, 
                   highlight=hl,
                   ax=ax2, clip_on=clip_on)
@@ -304,6 +366,10 @@ def gene_plot(chrom, start, end, highlight=[], hg19=False, only_protein_coding=F
 
     ax1.set_xlim(ax2.get_xlim())
 
+    if despine:
+        ax1.spines['top'].set_visible(False)
+        ax1.spines['right'].set_visible(False)
+
     return ax1
 
 
@@ -311,7 +377,7 @@ def gene_plot(chrom, start, end, highlight=[], hg19=False, only_protein_coding=F
 # Map between assembly coordinates
 ##################################################################################
 
-def map_interval(chrom, start, end, strand, map_from, map_to, species='human'):
+def map_interval(chrom, start, end, strand, map_from, map_to, species='homo_sapiens'):
     if chrom.startswith('chr'):
         chrom = chrom[3:]
     start, end = int(start), int(end)    
@@ -535,8 +601,12 @@ def _cached_symbol2ncbi(symbols, taxid=9606):
         for symbol in symbols:
             try:
                 geneids.append(symbol2ncbi.loc[symbol])
-            except:
-                print(f'Could not map "{symbol}" to ncbi id', file=sys.stderr)
+            except KeyError:
+                try:
+                    geneids.append(symbol2ncbi.loc[hgcn_symbol(symbol)])
+                    # geneids.append(ensembl2ncbi(ensembl_id(symbol)))
+                except NotFound:
+                    print(f'Could not map "{symbol}" to ncbi id', file=sys.stderr)
         return geneids
 
 
@@ -551,7 +621,7 @@ def _cached_ncbi2symbol(geneids, taxid=9606):
         for geneid in geneids:
             try:
                 symbols.append(ncbi2symbol.loc[geneid])
-            except:
+            except KeyError:
                 print(f'Could not map "{geneid}" to gene symbol', file=sys.stderr)
         return symbols
 
@@ -569,6 +639,13 @@ def _tidy_taxid(taxid):
     return taxid  
     
 
+def symbols_protein_coding(taxid=9606):
+    fetch_background_genes(taxid=taxid)
+    symbol2ncbi_file = f'geneinfo_cache/{taxid}_symbol2ncbi.h5'
+    symbol2ncbi = pd.read_hdf(symbol2ncbi_file, 'ncbi2symbol')
+    return symbol2ncbi.tolist()
+
+
 def get_terms_for_go_regex(regex, taxid=9606, add_children=False):
 
     taxid = _tidy_taxid(taxid)
@@ -781,16 +858,13 @@ def show_go_dag_for_terms(terms, add_relationships=True):
 # https://github.com/tanghaibao/goatools/blob/main/notebooks/goea_nbt3102_group_results.ipynb
 
 
-
-
-
-def show_go_dag_for_gene(gene, taxid=9606, evidence=None):
+def show_go_dag_for_gene(gene, taxid=9606, evidence=None, add_relationships=True):
     # evidence codes: http://geneontology.org/docs/guide-go-evidence-codes/
     go_terms = get_go_terms_for_genes([gene], taxid=taxid, evidence=evidence)
     if not go_terms:
         print('No GO terms to show', file=sys.stderr)
         return
-    return show_go_dag_for_terms(go_terms)
+    return show_go_dag_for_terms(go_terms, add_relationships=add_relationships)
 
     
 def show_go_evidence_codes():   
@@ -973,7 +1047,8 @@ class My_GOEnrichemntRecord(GOEnrichmentRecord):
         return f'<{self.GO}>'
 
 
-def go_enrichment(gene_list, taxid=9606, background_chrom=None, background_genes=None, terms=None, list_study_genes=False):
+def go_enrichment(gene_list, taxid=9606, background_chrom=None, background_genes=None, 
+    terms=None, list_study_genes=False, alpha=0.05):
 
     if type(gene_list) is pd.core.series.Series:
         gene_list = gene_list.tolist()
@@ -986,6 +1061,10 @@ def go_enrichment(gene_list, taxid=9606, background_chrom=None, background_genes
     
     taxid = _tidy_taxid(taxid)
 
+    ncbi_tsv = f'geneinfo_cache/{taxid}_protein_genes.txt'
+    if not os.path.exists(ncbi_tsv):
+        fetch_background_genes(taxid)
+
     with open(os.devnull, 'w') as null, redirect_stdout(null):
 
         obo_fname = download_and_move_go_basic_obo(prt=null)
@@ -1075,14 +1154,168 @@ def go_enrichment(gene_list, taxid=9606, background_chrom=None, background_genes
         .sort_values(by=['p_fdr_bh', 'ratio'])
         .reset_index(drop=True)
         )
-        return df.loc[df.p_fdr_bh < 0.05]
+        return df.loc[df.p_fdr_bh < alpha]
 
 
 
-def plot_go_enrichment_results(results):
-    
+def show_go_dag_enrichment_results(results):
+
     if type(results) is pd.core.series.Series:
         results = results.tolist()
     with open(os.devnull, 'w') as null, redirect_stdout(null):
         plot_results('geneinfo_cache/plot.png', results)
-    return Image('geneinfo_cache/plot.png')    
\ No newline at end of file
+    return Image('geneinfo_cache/plot.png')    
+
+
+def chrom_ideogram(annot, hspace=0.1, min_visible_width=200000, figsize=(10,10), assembly='hg38'):
+
+    d = {'axes.linewidth': 0.8, 'grid.linewidth': 0.64, 'lines.linewidth': 0.96, 
+         'lines.markersize': 3.84, 'patch.linewidth': 0.64, 'xtick.major.width': 0.8,
+         'ytick.major.width': 0.8, 'xtick.minor.width': 0.64, 'ytick.minor.width': 0.64,
+         'xtick.major.size': 3.84, 'ytick.major.size': 3.84, 'xtick.minor.size': 2.56, 
+         'ytick.minor.size': 2.56, 'font.size': 7.68, 'axes.labelsize': 7.68,
+         'axes.titlesize': 7.68, 'xtick.labelsize': 7.04, 'ytick.labelsize': 7.04, 
+         'legend.fontsize': 7.04, 'legend.title_fontsize': 7.68}
+    
+    chrom_lengths = {'hg19': {'chr1': 249250621, 'chr2': 243199373, 'chr3': 198022430, 'chr4': 191154276, 
+                              'chr5': 180915260, 'chr6': 171115067, 'chr7': 159138663, 'chr8': 146364022, 
+                              'chr9': 141213431, 'chr10': 135534747, 'chr11': 135006516, 'chr12': 133851895,
+                              'chr13': 115169878, 'chr14': 107349540, 'chr15': 102531392, 'chr16': 90354753, 
+                              'chr17': 81195210, 'chr18': 78077248, 'chr19': 59128983, 'chr20': 63025520, 
+                              'chr21': 48129895, 'chr22': 51304566, 'chrX': 155270560, 'chrY': 59373566},
+                     'hg38': {'chr1': 248956422, 'chr2': 242193529, 'chr3': 198295559, 'chr4': 190214555, 
+                              'chr5': 181538259, 'chr6': 170805979, 'chr7': 159345973, 'chr8': 145138636, 
+                              'chr9': 138394717, 'chr10': 133797422, 'chr11': 135086622, 'chr12': 133275309, 
+                              'chr13': 114364328, 'chr14': 107043718, 'chr15': 101991189, 'chr16': 90338345, 
+                              'chr17': 83257441, 'chr18': 80373285, 'chr19': 58617616, 'chr20': 64444167, 
+                              'chr21': 46709983, 'chr22': 50818468, 'chrX': 156040895, 'chrY': 57227415}}    
+
+    centromeres = {
+        'chr1':    (121700000, 125100000),
+        'chr10':   (38000000, 41600000),
+        'chr11':   (51000000, 55800000),
+        'chr12':   (33200000, 37800000),
+        'chr13':   (16500000, 18900000),
+        'chr14':   (16100000, 18200000),
+        'chr15':   (17500000, 20500000),
+        'chr16':   (35300000, 38400000),
+        'chr17':   (22700000, 27400000),
+        'chr18':   (15400000, 21500000),
+        'chr19':   (24200000, 28100000),
+        'chr2':    (91800000, 96000000),
+        'chr20':   (25700000, 30400000),
+        'chr21':   (10900000, 13000000),
+        'chr22':   (13700000, 17400000),
+        'chr3':    (87800000, 94000000),
+        'chr4':    (48200000, 51800000),
+        'chr5':    (46100000, 51400000),
+        'chr6':    (58500000, 62600000),
+        'chr7':    (58100000, 62100000),
+        'chr8':    (43200000, 47200000),
+        'chr9':    (42200000, 45500000),
+        'chrX':    (58100000, 63800000),
+        'chrY':    (10300000, 10400000)}                     
+
+    chr_names = [f'chr{x}' for x in list(range(1, 23))+['X', 'Y']]
+    chr_sizes = [chrom_lengths[assembly][chrom] for chrom in chr_names]
+    figwidth = max(chr_sizes)
+    
+    with plt.rc_context(d):
+    
+        nr_rows, nr_cols = len(chr_names)-2, 2
+
+        fig = plt.figure(figsize=figsize)
+
+        gs = matplotlib.gridspec.GridSpec(nr_rows, 25)
+        gs.update(wspace=0, hspace=hspace) # set the spacing between axes.             
+        ax_list = [plt.subplot(gs[i, :]) for i in range(nr_rows-2)]
+        ax_list.append(plt.subplot(gs[nr_rows-2, :9]))
+        ax_list.append(plt.subplot(gs[nr_rows-1, :9]))
+        ax_list.append(plt.subplot(gs[nr_rows-2, 9:]))
+        ax_list.append(plt.subplot(gs[nr_rows-1, 9:]))
+
+        chr_axes = dict(zip(chr_names, ax_list))
+
+        for ax in ax_list[:-4]:
+            ax.set_xlim((-200000, figwidth+100000))
+        for ax in ax_list[-4:]:
+            ax.set_xlim((-200000, ((25-9)/25)*figwidth+100000))
+
+        for i in range(len(ax_list)):
+            chrom = chr_names[i]
+            ax = ax_list[i]
+            start, end = 0, chr_sizes[i] 
+            ax.spines['top'].set_visible(False)
+            ax.spines['right'].set_visible(False)
+            ax.spines['bottom'].set_visible(False)
+            ax.spines['left'].set_visible(False)
+            ax.set_ylim((0, 3))
+
+            if i in [20, 21]:   
+                x = -3500000
+            else:
+                x = -2000000
+            ax.text(x, 1, chrom.replace('chr', ''), fontsize=8, horizontalalignment='right', weight='bold')
+
+            # h = ax.set_ylabel(chrom)
+            # h.set_rotation(0)
+            ax.set_yticklabels([])
+
+            if i == 0:
+                ax.spines['top'].set_visible(True)
+                ax.xaxis.tick_top()
+                ax.xaxis.set_label_position('top') 
+                ax.yaxis.set_ticks_position('none')
+            elif i == len(ax_list)-1:
+                ax.xaxis.tick_bottom()
+                ax.spines['bottom'].set_visible(True)                    
+                ax.yaxis.set_ticks_position('none')
+            else:
+                ax.set_xticklabels([])
+                ax.xaxis.set_ticks_position('none')
+                ax.yaxis.set_ticks_position('none')
+
+
+            # draw chrom
+            g = ax.add_patch(Rectangle((start, 1), end-start, 1, 
+                                       fill=False,
+                                       color='black',
+                                       edgecolor=None,
+                                       zorder=1, linewidth=0.7
+                                      ))
+
+            # draw centromere
+            cent_start, cent_end = centromeres[chrom]
+            ax.add_patch(Rectangle((cent_start, 0), cent_end-cent_start, 3, 
+                                       fill=True, color='white',
+                                       zorder=2))
+            xy = [[cent_start, 1], [cent_start, 2], [cent_end, 1], [cent_end, 2]]
+            g = ax.add_patch(Polygon(xy, closed=True, zorder=3, fill=True,
+                                     # color='#666666',
+                                     color='#777777',
+                                    ))
+
+
+        def plot_segment(chrom, start, end, color='red', label=None, base=1, height=1):
+            x, y, width = start, base, end-start
+
+            if width < min_visible_width:
+                x -= min_visible_width/2
+                width += min_visible_width
+
+            rect = Rectangle((x, y), width, height, linewidth=1, edgecolor='none', facecolor=color)
+            chr_axes[chrom].add_patch(rect)    
+            if label is not None:
+                chr_axes[chrom].plot([x+width/2, x+width/2], [y+height, y+height+0.5], linewidth=0.5, color=color)
+                chr_axes[chrom].text(x+width/2, y+height+0.5, label, fontsize=4, horizontalalignment='left',# weight='bold',
+                         verticalalignment='bottom', rotation=45)
+
+        for tup in annot:
+            plot_segment(*tup)                     
+            
+
+# annot = [('chr1', 20000000, 20100000, 'red', 'TP53'), ('chr7', 20000000, 30000000, 'orange', 'DYNLT3')] \
+# + [('chr5', 40000000, 70000000, 'red', None, 1, 0.5), ('chr8', 90000000, 110000000)] \
+#  + [('chrX', x[0], x[1], 'black', str(x[2]/1000000)) for x in zip(range(0, 150000000, 10000000), range(300000, 150000000, 10000000), range(0, 150000000, 10000000))]
+
+# chrom_ideogram(annot, figsize=(15, 9), hspace=0.2) 
\ No newline at end of file
